@@ -43,6 +43,7 @@
 #include "atmodem.h"
 
 static const char *csca_prefix[] = { "+CSCA:", NULL };
+static const char *cgsms_prefix[] = { "+CGSMS:", NULL };
 static const char *csms_prefix[] = { "+CSMS:", NULL };
 static const char *cmgf_prefix[] = { "+CMGF:", NULL };
 static const char *cpms_prefix[] = { "+CPMS:", NULL };
@@ -115,8 +116,7 @@ static void at_csca_set(struct ofono_sms *sms,
 		return;
 
 error:
-	if (cbd)
-		g_free(cbd);
+	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, user_data);
 }
@@ -181,8 +181,7 @@ static void at_csca_query(struct ofono_sms *sms, ofono_sms_sca_query_cb_t cb,
 		return;
 
 error:
-	if (cbd)
-		g_free(cbd);
+	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, NULL, user_data);
 }
@@ -245,8 +244,90 @@ static void at_cmgs(struct ofono_sms *sms, unsigned char *pdu, int pdu_len,
 		return;
 
 error:
-	if (cbd)
-		g_free(cbd);
+	g_free(cbd);
+
+	CALLBACK_WITH_FAILURE(cb, -1, user_data);
+}
+
+static void at_cgsms_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	ofono_sms_sca_set_cb_t cb = cbd->cb;
+	struct ofono_error error;
+
+	decode_at_error(&error, g_at_result_final_response(result));
+
+	cb(&error, cbd->data);
+}
+
+static void at_cgsms_set(struct ofono_sms *sms, int bearer,
+			ofono_sms_bearer_set_cb_t cb, void *user_data)
+{
+	struct sms_data *data = ofono_sms_get_data(sms);
+	struct cb_data *cbd = cb_data_new(cb, user_data);
+	char buf[64];
+
+	if (!cbd)
+		goto error;
+
+	snprintf(buf, sizeof(buf), "AT+CGSMS=%d", bearer);
+
+	if (g_at_chat_send(data->chat, buf, none_prefix,
+				at_cgsms_set_cb, cbd, g_free) > 0)
+		return;
+
+error:
+	g_free(cbd);
+
+	CALLBACK_WITH_FAILURE(cb, user_data);
+}
+
+static void at_cgsms_query_cb(gboolean ok, GAtResult *result,
+				gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	ofono_sms_bearer_query_cb_t cb = cbd->cb;
+	struct ofono_error error;
+	GAtResultIter iter;
+	int bearer;
+
+	decode_at_error(&error, g_at_result_final_response(result));
+
+	if (!ok) {
+		cb(&error, -1, cbd->data);
+		return;
+	}
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "+CGSMS:"))
+		goto err;
+
+	g_at_result_iter_next_number(&iter, &bearer);
+
+	cb(&error, bearer, cbd->data);
+
+	return;
+
+err:
+	CALLBACK_WITH_FAILURE(cb, -1, cbd->data);
+}
+
+static void at_cgsms_query(struct ofono_sms *sms,
+				ofono_sms_bearer_query_cb_t cb, void *user_data)
+{
+	struct sms_data *data = ofono_sms_get_data(sms);
+	struct cb_data *cbd = cb_data_new(cb, user_data);
+
+	if (!cbd)
+		goto error;
+
+	if (g_at_chat_send(data->chat, "AT+CGSMS?", cgsms_prefix,
+				at_cgsms_query_cb, cbd, g_free) > 0)
+		return;
+
+error:
+	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, -1, user_data);
 }
@@ -284,6 +365,8 @@ static inline void at_ack_delivery(struct ofono_sms *sms)
 	struct sms_data *data = ofono_sms_get_data(sms);
 	char buf[256];
 
+	DBG("");
+
 	/* We must acknowledge the PDU using CNMA */
 	if (data->cnma_ack_pdu)
 		snprintf(buf, sizeof(buf), "AT+CNMA=1,%d\r%s",
@@ -297,6 +380,7 @@ static inline void at_ack_delivery(struct ofono_sms *sms)
 static void at_cds_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	long pdu_len;
 	int tpdu_len;
 	const char *hexpdu;
@@ -318,7 +402,8 @@ static void at_cds_notify(GAtResult *result, gpointer user_data)
 	decode_hex_own_buf(hexpdu, -1, &pdu_len, 0, pdu);
 	ofono_sms_status_notify(sms, pdu, pdu_len, tpdu_len);
 
-	at_ack_delivery(sms);
+	if (data->cnma_enabled)
+		at_ack_delivery(sms);
 }
 
 static void at_cmt_notify(GAtResult *result, gpointer user_data)
@@ -357,6 +442,8 @@ static void at_cmgr_notify(GAtResult *result, gpointer user_data)
 	long pdu_len;
 	int tpdu_len;
 
+	DBG("");
+
 	g_at_result_iter_init(&iter, result);
 
 	if (!g_at_result_iter_next(&iter, "+CMGR:"))
@@ -505,6 +592,8 @@ static void at_cmgl_done(struct ofono_sms *sms)
 {
 	struct sms_data *data = ofono_sms_get_data(sms);
 
+	DBG("");
+
 	if (data->incoming == AT_UTIL_SMS_STORE_MT &&
 			data->store == AT_UTIL_SMS_STORE_ME) {
 		at_cmgl_set_cpms(sms, AT_UTIL_SMS_STORE_SM);
@@ -538,6 +627,8 @@ static void at_cmgl_notify(GAtResult *result, gpointer user_data)
 	int status;
 	char buf[16];
 
+	DBG("");
+
 	g_at_result_iter_init(&iter, result);
 
 	while (g_at_result_iter_next(&iter, "+CMGL:")) {
@@ -705,6 +796,8 @@ static gboolean build_cnmi_string(char *buf, int *cnmi_opts,
 	const char *mode;
 	int len = sprintf(buf, "AT+CNMI=");
 
+	DBG("");
+
 	if (data->vendor == OFONO_VENDOR_QUALCOMM_MSM ||
 			data->vendor == OFONO_VENDOR_HUAWEI ||
 			data->vendor == OFONO_VENDOR_NOVATEL)
@@ -734,14 +827,14 @@ static gboolean build_cnmi_string(char *buf, int *cnmi_opts,
 	 * sends the device into la-la land.
 	 */
 	if (data->vendor == OFONO_VENDOR_NOVATEL)
-		/* MSM devices advertise support for mode 2, but return an
-		 * error if we attempt to actually use it. */
-		mode = "2";
+		mode = "20";
 	else
-		/* Sounds like 2 is the sanest mode */
-		mode = data->cnma_enabled ? "10" : "20";
+		mode = "120";
 
-	/* Always deliver Status-Reports via +CDS or don't deliver at all */
+	/*
+	 * Try to deliver Status-Reports via +CDS, then CDSI or don't
+	 * deliver at all
+	 * */
 	if (!append_cnmi_element(buf, &len, cnmi_opts[3], mode, FALSE))
 		return FALSE;
 
@@ -759,6 +852,8 @@ static void construct_ack_pdu(struct sms_data *d)
 	int len;
 	int tpdu_len;
 
+	DBG("");
+
 	memset(&ackpdu, 0, sizeof(ackpdu));
 
 	ackpdu.type = SMS_TYPE_DELIVER_REPORT_ACK;
@@ -822,8 +917,9 @@ static void at_cnmi_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		supported = TRUE;
 
 	/* support for ack pdu is not working */
-	if (data->vendor == OFONO_VENDOR_NOVATEL ||
+	if (data->vendor == OFONO_VENDOR_IFX ||
 			data->vendor == OFONO_VENDOR_HUAWEI ||
+			data->vendor == OFONO_VENDOR_NOVATEL ||
 			data->vendor == OFONO_VENDOR_OPTION_HSO)
 		goto out;
 
@@ -1124,12 +1220,12 @@ static int at_sms_probe(struct ofono_sms *sms, unsigned int vendor,
 	struct sms_data *data;
 
 	data = g_new0(struct sms_data, 1);
-	data->chat = chat;
+	data->chat = g_at_chat_clone(chat);
 	data->vendor = vendor;
 
 	ofono_sms_set_data(sms, data);
 
-	g_at_chat_send(chat, "AT+CSMS=?", csms_prefix,
+	g_at_chat_send(data->chat, "AT+CSMS=?", csms_prefix,
 			at_csms_query_cb, sms, NULL);
 
 	return 0;
@@ -1139,12 +1235,12 @@ static void at_sms_remove(struct ofono_sms *sms)
 {
 	struct sms_data *data = ofono_sms_get_data(sms);
 
-	if (data->cnma_ack_pdu)
-		g_free(data->cnma_ack_pdu);
+	g_free(data->cnma_ack_pdu);
 
 	if (data->timeout_source > 0)
 		g_source_remove(data->timeout_source);
 
+	g_at_chat_unref(data->chat);
 	g_free(data);
 }
 
@@ -1155,6 +1251,8 @@ static struct ofono_sms_driver driver = {
 	.sca_query	= at_csca_query,
 	.sca_set	= at_csca_set,
 	.submit		= at_cmgs,
+	.bearer_query	= at_cgsms_query,
+	.bearer_set	= at_cgsms_set,
 };
 
 void at_sms_init()
